/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Remy Mahler, digitalSTROM AG <remy.mahler@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "channelutil.h"
#include "datamodel.h"
#include "util.h"

typedef struct OPCChannelConversionTag {
    double channelMinVal;
    double channelMaxVal;
    uint8_t channelSize; // 8 bit
} OPCChannelConversion;

/* TABLE for opc channel definition */
/* see also: https://intranet.aizo.net/Wiki/Output%20Channel%20Types.aspx */
static OPCChannelConversion OpcChannelRange[] =
{
/*   channel    min     max         1:8 bit 2:16 bit */
    { /*1*/     0.0,    100.0,      1},
    { /*2*/     0.0,    358.6,      1},
    { /*3*/     0.0,    100.0,      1},
    { /*4*/     100.0,  1000.0,     1},
    { /*5*/     0.0,    1.0,        1},
    { /*6*/     0.0,    1.0,        1},
    { /*7*/     0.0,    100.0,      2},
    { /*8*/     0.0,    100.0,      2},
    { /*9*/     0.0,    100.0,      1},
    { /*10*/    0.0,    100.0,      1}
};

#define MAX_OPC_CHANNELS  (sizeof(OpcChannelRange)/sizeof(OpcChannelRange[0]))
#define MAX_VAL_8 255
#define MAX_VAL_16 65535

static int translate_lookupTableIndex(int index)
{
    if (index > 0) {
        --index;
    }
    return index;
}

static bool check_valid_channel( uint8_t channel)
{
    if (channel >= MAX_OPC_CHANNELS) {
        vdsm_report(LOG_ERR, "util convert dsmval to vdcapival: channel id:%i is out of range\n", channel);
        return false;
    }
    return true;
}

uint16_t convert_vcdapival_to_dsmval(uint8_t channelId,
                                     double vdcapivalue, uint16_t *dsmval)
{
    if (!check_valid_channel(channelId)) {
        vdsm_report(LOG_ERR, "util convert vdcapival to dsmval : channel id:%i is out of range\n", channelId);
        return false;
    }

    bool is16Bit = (OpcChannelRange[translate_lookupTableIndex(channelId)].channelSize == 2);
    double max = OpcChannelRange[translate_lookupTableIndex(channelId)].channelMaxVal;
    double min = OpcChannelRange[translate_lookupTableIndex(channelId)].channelMinVal;

    if (is16Bit) {
        *dsmval = (uint16_t)((vdcapivalue-min)/(max-min) * MAX_VAL_16);
    } else {
        *dsmval = (uint16_t)((vdcapivalue-min)/(max-min) * MAX_VAL_8);
    }
    return true;
}

bool convert_dsmval_to_vdcapival(uint8_t channelId,
                                 uint16_t value, double *vDCAPIValue)
{
    if (!check_valid_channel(channelId)) {
        vdsm_report(LOG_ERR, "util convert dsmval to vdcapival : channel id: %i is out of range\n", channelId);
        return false;
    }

    bool is16Bit = (OpcChannelRange[translate_lookupTableIndex(channelId)].channelSize == 2);
    double max = OpcChannelRange[translate_lookupTableIndex(channelId)].channelMaxVal;
    double min = OpcChannelRange[translate_lookupTableIndex(channelId)].channelMinVal;

    if (is16Bit) {
        *vDCAPIValue = (value*(max-min)/MAX_VAL_16)+min;
    } else {
        *vDCAPIValue = (value*(max-min)/MAX_VAL_8)+min;
    }
    return true;
}

bool util_check_valid_params(uint8_t channelId, double minVal, double maxVal)
{
    if (channelId >= MAX_OPC_CHANNELS) return false;

    return ( (maxVal == OpcChannelRange[translate_lookupTableIndex(channelId)].channelMaxVal) &&
             (minVal == OpcChannelRange[translate_lookupTableIndex(channelId)].channelMinVal) );
}

