/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include <boost/asio/steady_timer.hpp>
#include "ds485.h"

#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include "bsp.h"

extern "C" {
#include <digitalSTROM/ds485-socket-server.h>
#include <digitalSTROM/ds485-socket-interface.h>

#include "dsmapi.h"
#include "dsm.h"
#include "datamodel.h"
#include "mmi.h"
#include "bitset8.h"
#include "types.h"
#include "util.h"
#include "vdc/vdc-api-proto.h"
} //extern "C"

#define LOG(_loglevel, _message, ...) vdsm_report(_loglevel, _message, ##__VA_ARGS__)
typedef boost::system::error_code error_code;

Q_DEFINE_THIS_MODULE(ds485)

typedef struct dS485Tag {
    QActive super;
    dsuid_t dsmid;
    bool terminate;
    bool listenGlobal;
    const char* listenPort;
    pthread_t rxThread;
    struct dS485_server* netlibHandle;
} DS485;

/* Active object class -----------------------------------------------------*/

static QState DS485_initial (DS485 *me, QEvent const *e);
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e);
static QState DS485_top (DS485 *me, QEvent const *e);
static QState DS485_startup(DS485 *me, QEvent const *e);

static QState DS485_operational(DS485 *me, QEvent const *e);

static QState DS485_stopped(DS485 *me, QEvent const *e);

static void* rxThread(void* arg);

/* Local objects -----------------------------------------------------------*/
static DS485 l_ds485;   /* the single instance of the dS485 active object */
static QEvent const * l_ds485TxMessageQueueSto[128];    /* event queue storage */
static const QEvent l_startupEnableEvt = { DS485_RX_STARTUP_SIG, 0 };
static QEQueue l_ds485TxMessageQueue;   /* transmit message queue */
static boost::asio::io_service* l_ioService;


/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DS485 = (QActive *)&l_ds485;      /* "opaque" AO pointer */

class LongAction;

/** Maps vdc message ids to vdc pending replies and guards their timeout. */
class LongActions {
public:
    LongActions(boost::asio::io_service& ioService);
    void register_(std::shared_ptr<LongAction> action);
    bool processLongReply(int messageId, const unsigned char *data, const uint16_t len);

private:
    std::map<int, std::shared_ptr<LongAction>> m_map; // TODO(c++11): use std::unique_ptr
    boost::asio::basic_waitable_timer<boost::chrono::steady_clock> m_timeoutTimer;
    bool m_timeoutTimerActive;

    void startTimeoutTimer();
};

static LongActions* l_longActions;

struct VdcApiFreeMessageDeleter {
    void operator()(Vdcapi__Message *msg) { vdc_api_free_message(msg); }
};

class LongAction {
public:
    static void processPacket(const ds485n_packet_t& packet) {
        std::shared_ptr<LongAction> action(new LongAction(packet));
        try {
            if (action->m_messageId != USER_PROTOBUF_MESSAGE_REQUEST) {
                throw std::runtime_error("Wrong message id");
            }

            char destinationString[DSUID_STR_LEN];
            dsuid_to_string_upper(&action->m_destinationDsuid, destinationString);
            VDSM_INFO("received long packet messageId:%d destination:%s", action->m_messageId, destinationString);

            vdc_client_t *vdcClient;
            if (!vdc_api_get_vdc_host_for_vdc(destinationString, &vdcClient)) {
                throw std::runtime_error(std::string("Vdc client not registered ") + destinationString);
            }

            std::unique_ptr<Vdcapi__Message, VdcApiFreeMessageDeleter> message;
            {
                uint16_t size = (packet.raw_data[DSMAPI_LONG_REQUEST_HEADER_LEN + 0] << 8)
                        | packet.raw_data[DSMAPI_LONG_REQUEST_HEADER_LEN + 1];
                const uint8_t* data = &packet.raw_data[DSMAPI_LONG_REQUEST_HEADER_LEN + 2];
                if (size > VDC_USER_PROTOBUF_MESSAGE_LENGTH) {
                    throw std::runtime_error("Data size too big");
                }
                message.reset(vdcapi__message__unpack(NULL, size, data));
                if (!message) {
                    throw std::runtime_error("Failed to unpack incoming message\n");
                }
            }

            message->message_id = action->m_vdcMessageId;
            message->has_message_id = 1;

            auto messageDataSize = vdcapi__message__get_packed_size(message.get());
            std::unique_ptr<uint8_t[]> messageData(new uint8_t[messageDataSize]);
            vdcapi__message__pack(message.get(), messageData.get());

            protobufMessagePrint(&message->base, 0);
            AsioVdcClient_asyncWriteFrame(vdcClient, messageData.get(), messageDataSize);
        } catch (const std::runtime_error &e) {
            VDSM_ERROR("%s", e.what());
            action->replyError(ERROR_WRONG_PARAMETER);
            return;
        }
        l_longActions->register_(std::move(action));
    }

    void replyOk(const void *data, uint16_t dataLength) {
        replyImpl(ERROR_OK, data, dataLength);
    }

    void replyError(int error) {
        replyImpl(error, /* nullptr */ 0, 0);
    }

    void replyTimeoutError() {
        replyError(ERROR_TIMEOUT);
    }

    int getVdcMessageId() const { return m_vdcMessageId; }
    boost::chrono::time_point<boost::chrono::steady_clock> getTimeoutTimePoint() const { return m_timeoutTimePoint; }

private:
    dsuid_t m_sourceDsuid;
    dsuid_t m_destinationDsuid;
    uint8_t m_transactionId;
    uint8_t m_messageId;
    uint8_t m_modifierId;
    int m_vdcMessageId;
    boost::chrono::time_point<boost::chrono::steady_clock> m_timeoutTimePoint;

    LongAction(const ds485n_packet_t& packet)
            : m_sourceDsuid(packet.sourceId)
            , m_destinationDsuid(packet.destinationId)
            , m_transactionId(packet.raw_data[DSMAPI_LONG_TRANSACTIONID_OFFSET - DS485N_PACKET_DATA_OFFSET])
            , m_messageId(packet.raw_data[DSMAPI_LONG_MSGID_OFFSET - DS485N_PACKET_DATA_OFFSET])
            , m_modifierId(packet.raw_data[DSMAPI_LONG_MODIFIER_OFFSET - DS485N_PACKET_DATA_OFFSET])
            , m_vdcMessageId(vdc_api_next_message_id())
            , m_timeoutTimePoint(boost::chrono::steady_clock::now() + boost::chrono::seconds(5))
    {
    }

    void replyImpl(uint8_t error, const void *data, uint16_t dataLength) {
        VDSM_IENTER("messageId:%d error:%d dataLength:%d", m_messageId, error, dataLength);
        if (dsuid_is_broadcast(&m_destinationDsuid)) {
            m_destinationDsuid = Datamodel_GetDsmid();
        }

        ds485n_packet_t packet;
        memset(&packet, 0, sizeof(packet));
        packet.containerType = DS485_CONTAINER_RESPONSE;
        packet.destinationId = m_sourceDsuid;
        packet.sourceId = m_destinationDsuid;
        packet.length = DSMAPI_LONG_RESPONSE_HEADER_LEN + 2 + dataLength;
        packet.raw_data[DSMAPI_LONG_TRANSACTIONID_OFFSET - DS485N_PACKET_DATA_OFFSET] = m_transactionId;
        packet.raw_data[DSMAPI_LONG_MSGID_OFFSET - DS485N_PACKET_DATA_OFFSET] = m_messageId;
        packet.raw_data[DSMAPI_LONG_MODIFIER_OFFSET - DS485N_PACKET_DATA_OFFSET] = m_modifierId;
        packet.raw_data[DSMAPI_LONG_ERROR_OFFSET - DS485N_PACKET_DATA_OFFSET] = error;
        packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 0] = uint8_t(dataLength >> 8); //dsm-api payload starts here
        packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 1] = uint8_t(dataLength >> 0);
        memcpy(&packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 2], data, dataLength);
        ds485_server_reply_long(l_ds485.netlibHandle, &packet);
    }
};

LongActions::LongActions(boost::asio::io_service &ioService) : m_timeoutTimer(ioService), m_timeoutTimerActive(false) {}

void LongActions::register_(std::shared_ptr<LongAction> action) {
    auto emplaceRet = m_map.insert(std::make_pair(action->getVdcMessageId(), std::move(action)));
    Q_ASSERT(emplaceRet.second); //This can happen if we receive more than 4G requests per timeout
    startTimeoutTimer();
}

bool LongActions::processLongReply(int messageId, const unsigned char *data, const uint16_t len) {
    auto it = m_map.find(messageId);
    if (it == m_map.end()) {
        return false;
    }
    try {
        it->second->replyOk(data, len);
    } catch (const std::runtime_error &e) {
        VDSM_ERROR("%s", e.what());
    }
    VDSM_ENTER("m_map.size:%d", m_map.size());
    m_map.erase(it);
    return true;
}

void LongActions::startTimeoutTimer() {
    if (m_timeoutTimerActive) {
        return;
    }
    if (m_map.empty()) {
        return;
    }
    VDSM_ENTER("m_map.size:%d", m_map.size());

    m_timeoutTimer.expires_from_now(boost::chrono::seconds(5));
    m_timeoutTimer.async_wait([this](const error_code& e) {
        if (e) {
            return; //timer was aborted
        }
        VDSM_ENTER("m_map.size:%d", m_map.size());

        m_timeoutTimerActive = false;
        auto now = boost::chrono::steady_clock::now();
        for (auto it = m_map.begin(); it != m_map.end(); ) {
            auto& action = *it->second;
            if (action.getTimeoutTimePoint() <= now) {
                action.replyTimeoutError();
                it = m_map.erase(it);
            } else {
                it++;
            }
        }
        startTimeoutTimer();
    });
    m_timeoutTimerActive = true;
}

int DS485_processLongReply(int messageId, const unsigned char *data, const uint16_t len) {
    return l_longActions->processLongReply(messageId, data, len);
}

/*..........................................................................*/
void DS485_ctor(boost::asio::io_service& ioService, const char* listenPort, bool listenGlobal)
{
    DS485 *me = &l_ds485;
    QActive_ctor(&me->super, (QStateHandler)&DS485_initial);
    /* initialize message event transmit queue */
    QEQueue_init(&l_ds485TxMessageQueue, l_ds485TxMessageQueueSto, Q_DIM(l_ds485TxMessageQueueSto), "ds485TxMessage", &me->super);
    /* initialize time events */
    me->terminate = FALSE;
    me->listenGlobal = listenGlobal;
    me->listenPort = listenPort;
    l_ioService = &ioService;
    l_longActions = new LongActions(ioService);
}

uint8_t DS485_GetTxMessageQueueMargin(void)
{
    return QEQueue_getNFree(&l_ds485TxMessageQueue);
}

uint8_t DS485_GetDeferredTxMessageQueueMargin(void)
{
    return 0;
}

void DS485_SendUserPropertyEvent(const dsuid_t* sourceDsm, const uint8_t* data, const uint16_t dataLength)
{
    Q_ASSERT(DS485N_PACKET_MAX_PAYLOAD_SIZE_RAW - DSMAPI_LONG_RESPONSE_HEADER_LEN >= dataLength + 2);
    ds485n_packet_t packet;
    memset(&packet, 0, sizeof(packet));
    packet.containerType = DS485_CONTAINER_EVENT;
    packet.destinationId = DSUID_BROADCAST;
    packet.sourceId = *sourceDsm;
    packet.length = DSMAPI_LONG_RESPONSE_HEADER_LEN + 2 + dataLength;
    packet.raw_data[DSMAPI_LONG_TRANSACTIONID_OFFSET - DS485N_PACKET_DATA_OFFSET] = 255;
    packet.raw_data[DSMAPI_LONG_MSGID_OFFSET - DS485N_PACKET_DATA_OFFSET] = EVENT_USER_PROPERTY; // command
    packet.raw_data[DSMAPI_LONG_MODIFIER_OFFSET - DS485N_PACKET_DATA_OFFSET] = 0; //modifier
    packet.raw_data[DSMAPI_LONG_ERROR_OFFSET - DS485N_PACKET_DATA_OFFSET] = ERROR_OK;
    packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 0] = uint8_t(dataLength >> 8); //dsm-api payload starts here
    packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 1] = uint8_t(dataLength >> 0);
    memcpy(&packet.raw_data[DSMAPI_LONG_RESPONSE_HEADER_LEN + 2], data, dataLength);
    ds485_server_reply_long(l_ds485.netlibHandle, &packet);
}

void DS485_EmitBusChange(const char *dsuid, uint8_t flags)
{
    dsuid_t vdcuid;
    dsuid_from_string(dsuid, &vdcuid);

    DS485ReqBusChangeEvt * msgEvt = Q_NEW(DS485ReqBusChangeEvt, DS485_RX_JOIN_SIG);
    msgEvt->dsuid = vdcuid;
    msgEvt->flags = flags;
    QActive_postFIFO(AO_DS485, (QEvent *)msgEvt);
}

/**
 * Create DS485 control message
 */
// not used in vdsm
DS485MessageEvt * DS485_CreateControlMessage(const dsuid_t *dest, uint8_t src, DS485CommandType type,
                                             DS485CommandId cmd, uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_SHORT_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = DS485_TYPE_CONTROL;
    message[DS485_SRC_OFFSET] = src | type;
    message[DS485_CL_OFFSET] = (cmd << 4) | (payloadLength & 0xF);
    msgEvt->destinationAddr = *dest;
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessageFromId(const dsuid_t *destDsuid, uint16_t sourceId, uint8_t src, DS485CommandType type, uint8_t payloadLength)
{
    dsuid_t source_dsuid;
    if (sourceId == 0) {
        /* use vdsm dsuid as source */
        source_dsuid = Datamodel_GetDsmid();
    } else if (!Datamodel_GetVdcDsuidFromDeviceId(sourceId, &source_dsuid)) {
        vdsm_report(LOG_DEBUG, "can not convert source id %d to dsuid_t\n", sourceId);
        source_dsuid = DSUID_NULL;
    }
    return DS485_CreateDataMessageFromSrcDestDsuid(destDsuid, &source_dsuid, src, type, payloadLength);
}

DS485MessageEvt * DS485_CreateDataMessageFromSrcDestDsuid(const dsuid_t *destDsuid, const dsuid_t *sourceDsuid, uint8_t src, DS485CommandType type, uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_LONG_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt;
    if (payloadLength <= DS485_SHORT_PAYLOAD_SIZE) {
        msgEvt = (DS485MessageEvt *) Q_NEW(DS485ShortMessageEvt, DS485_TX_MSG_SIG);
    } else {
        msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    }
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = src | type;
    message[DS485_CL_OFFSET] = payloadLength;
    msgEvt->destinationAddr = *destDsuid;
    msgEvt->sourceAddr = *sourceDsuid;

    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
DS485MessageEvt * DS485_CreateDataMessageToDsm(uint8_t broadcast,
                                               const dsuid_t *srcId,
                                               const dsuid_t *destinationId,
                                               DS485CommandType type,
                                               uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_LONG_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt;
        msgEvt = Q_NEW(DS485MessageEvt, DS485_RX_DATA_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = broadcast | DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = type;
    message[DS485_CL_OFFSET] = payloadLength;
    unaligned_memcpy((void *)&msgEvt->sourceAddr, (void *)srcId, sizeof(dsuid_t));
    unaligned_memcpy((void *)&msgEvt->destinationAddr, (void *)destinationId, sizeof(dsuid_t));
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Return next DS485 message to transmit or null
 */
DS485MessageEvt const * DS485_GetNextTxMessage(void)
{
    return (DS485MessageEvt const *)QEQueue_get(&l_ds485TxMessageQueue);
}

/*..........................................................................*/
static QState DS485_initial(DS485 *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
    return Q_TRAN(&DS485_wait_for_configuration);
}

/*..........................................................................*/
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case CONFIG_INIT_SIG: {
        DatamodelConfigEvt const * config = (DatamodelConfigEvt const *)e;
        /* save dSM ID */
        me->dsmid = config->dsmid;
        /* subscribe events */
        QActive_subscribe((QActive *)me, SYSTEM_TERM_SIG);
        return Q_TRAN(&DS485_startup);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_top(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case SYSTEM_TERM_SIG: {
        return Q_TRAN(&DS485_stopped);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_startup(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* startup receiver thread */
        if (pthread_create(&me->rxThread, NULL, rxThread, me)) {
            LOG(LOG_ERR, "Could not create receiver thread.\n");
            /* TODO: die */
        }
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        return Q_TRAN(&DS485_operational);
    }
//    case DS485_TX_MSG_SIG: {
//        /* defer transmit message event */
//        if (QEQueue_getNFree(&me->deferredTxMessageQueue) != 0) {
//            QActive_defer((QActive *)me, &me->deferredTxMessageQueue, e);
//        }
//        return Q_HANDLED();
//    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_operational(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485AddressEvt * addrEvt = Q_NEW(DS485AddressEvt, DS485_ADDRESS_SIG);
        addrEvt->deviceAddress = 0;
        QF_publish((QEvent *)addrEvt);
        return Q_HANDLED();
    }
    case DS485_TX_MSG_SIG: {
        DS485MessageEvt const * msgEvt = (DS485MessageEvt const *)e;
        ds485_container_t container;
        memset(&container, 0, sizeof(container));

        if (DS485_IsRequest(msgEvt)) {
            container.containerType = DS485_CONTAINER_REQUEST;
        } else if (DS485_IsResponse(msgEvt)) {
            container.containerType = DS485_CONTAINER_RESPONSE;
        } else {
            container.containerType = DS485_CONTAINER_EVENT;
        }

        container.destinationId = msgEvt->destinationAddr;
        container.sourceId = msgEvt->sourceAddr;

        container.length = DS485_GetPayloadLength(msgEvt) - 1;
        memcpy(container.raw_data, &msgEvt->message[DS485_DATA_OFFSET],
               container.length + 1);
        ds485_server_reply(me->netlibHandle, &container);
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        DS485ReqBusChangeEvt const * msgEvt = (DS485ReqBusChangeEvt const *)e;
        ds485_server_emit_bus_change(me->netlibHandle, msgEvt->dsuid, msgEvt->flags);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_stopped(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* wait till thread is terminated */
        int s;
        void *res;
        me->terminate = TRUE;
        ds485_server_abort(me->netlibHandle);
        s = pthread_join(me->rxThread, &res);
        if (s != 0) {
            LOG(LOG_ERR, "DS485_stopped: pthread_join returned error %d\n", s);
        }
        free(res);
        ds485_server_cleanup(me->netlibHandle);
        /* send DS485 stopped event */
        static const QEvent ds485StoppedEvt = { DS485_STOPPED_SIG, 0 };
        QActive_postFIFO(AO_MMI, &ds485StoppedEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

int ds485_server_send_data(struct dS485_server *server,
                           const ds485_container_t* packet)
{
    (void) server;
    //io_service::dispatch requires copyable functor -> shared pointer
    boost::packaged_task<int> task([=]() {
        DS485MessageEvt *evt;
        uint8_t broadcast;
        DS485 *me = &l_ds485;

        if (!(dsuid_is_broadcast(&packet->destinationId) ||
            vdc_api_vdsm_holds_vdc(&packet->destinationId) ||
            dsuid_equal(&me->dsmid, &packet->destinationId))) {
            return 0;
        }

        broadcast = dsuid_is_broadcast(&packet->destinationId) ?  DS485_BROADCAST_ADDRESS : 0;
        evt = DS485_CreateDataMessageToDsm(broadcast,
                                        &packet->sourceId,
                                        &packet->destinationId,
                                        DS485_TYPE_REQ,
                                        packet->length + 1);

        unaligned_memcpy((void *)&evt->message[DS485_DATA_OFFSET],
                        (void *)packet->raw_data, packet->length + 1);
        QActive_postFIFO(AO_DsmApi, (QEvent *)evt);
        return 0;
    });
    //dispatch task in io_service thread and block till it finishes.
    l_ioService->dispatch([&]() { task(); });
    return task.get_future().get();
}

int ds485_server_send_long_data(struct dS485_server *server,
                                const ds485n_packet_t* packet)
{
    (void) server;
    boost::packaged_task<int> task([=]() {
        try {
            if (dsuid_is_broadcast(&packet->destinationId) ||
                dsuid_equal(&l_ds485.dsmid, &packet->destinationId)) {
                return 0;
            }
            LongAction::processPacket(*packet);
        } catch (const std::runtime_error &e) {
            VDSM_ERROR("%s", e.what());
        }
        return 0;
    });
    l_ioService->dispatch([&]() { task(); });
    return task.get_future().get();
}

int ds485_server_get_bus_members(struct dS485_server *server,
                                 dsuid_t *devices, int length)
{
    (void) server;
    boost::packaged_task<int> task([=]() {
        int count = vdc_api_get_vdc_dsuid(devices, length);

        if (count < length) {
            DS485 *me = &l_ds485;
            memcpy(&devices[count], &(me->dsmid), sizeof(dsuid_t));
            ++count;
        }
        return count;
    });

    l_ioService->dispatch([&]() { task(); });
    return task.get_future().get();
}

dsuid_t ds485_server_get_dsuid(struct dS485_server *server)
{
    //BUG me->dsmid is empty until CONFIG_INIT_SIG is received.
    (void) server;

    boost::packaged_task<dsuid_t> task([]() {
        return l_ds485.dsmid;
    });
    l_ioService->dispatch([&]() { task(); });
    return task.get_future().get();
}

int ds485_server_restart_bus(struct dS485_server *server)
{
    (void) server;
    return 0;
}

int ds485_server_set_packetfilter(struct dS485_server *server, uint8_t rxfilter)
{
    (void) server;
    (void) rxfilter;
    return 0;
}

static void* rxThread(void* arg)
{
    DS485 *me = (DS485*)arg;

    struct dS485_server_callbacks netlib_cb;
    bzero(&netlib_cb, sizeof(netlib_cb));
    netlib_cb.receive_data = ds485_server_send_data;
    netlib_cb.receive_long_data = ds485_server_send_long_data;
    netlib_cb.get_bus_members = ds485_server_get_bus_members;
    netlib_cb.restart_bus = ds485_server_restart_bus;
    netlib_cb.get_dsuid = ds485_server_get_dsuid;
    netlib_cb.set_packetfilter = ds485_server_set_packetfilter;
    me->netlibHandle = ds485_server_init(&netlib_cb, false);
    if (me->netlibHandle == NULL) {
        LOG(LOG_ERR, "Could not allocate netlib handle.\n");
        return NULL;
    }
    l_ioService->dispatch([me]()
    {
        QActive_postFIFO((QActive*)me, &l_startupEnableEvt);
    });
    /* call netlib server listen function */
    ds485_server_listen(me->netlibHandle, strtol(me->listenPort, NULL, 10), me->listenGlobal);
    return NULL;
}
