/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "mmi.h"
#include "dsm.h"
#include "bsp.h"
#include "ds485.h"
#include "datamodel.h"
#include "dsmapi.h"

Q_DEFINE_THIS_MODULE(mmi)

/* Active object class -----------------------------------------------------*/
typedef struct MMITag {
    QActive super;
    QEQueue deferredQueue;
    QStateHandler flashHistory;
    MMIConfig config;
    bool ds485Stopped;
    bool datamodelStopped;
    bool terminate;
} MMI;

/* Local objects -----------------------------------------------------------*/
static MMI l_mmi;   /* the single instance of the mmi active object */
static QEvent const * l_deferredQueueSto[128];   /* event queue storage */
static const QEvent l_userAckEvt = { MMI_USER_ACK_SIG, 0 };

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_MMI = (QActive *)&l_mmi;      /* "opaque" AO pointer */

/* Local methods -----------------------------------------------------------*/
static QState MMI_initial(MMI *me, QEvent const *e);
static QState MMI_top(MMI *me, QEvent const *e);
static QState MMI_active(MMI *me, QEvent const *e);
static QState MMI_startup(MMI *me, QEvent const *e);
static QState MMI_shutdown(MMI *me, QEvent const *e);                   /* priority -1 (highest priority without exit) */
static QState MMI_invalid_configuration(MMI *me, QEvent const *e);      /* priority 3 */
static QState MMI_ready(MMI *me, QEvent const *e);                      /* priority 5 (lowest priority with exit) */
static QStateHandler MMI_DetermineState(MMI *me);

static void MMI_Shutdown(void);
static void MMI_SendChangeEvent(MMI *me);

/*..........................................................................*/
void MMI_ctor(void)
{
    MMI *me = &l_mmi;
    /* call active object (superclass) constructor */
    QActive_ctor(&me->super, (QStateHandler)&MMI_initial);
    /* initialize deferred event queue */
    QEQueue_init(&me->deferredQueue, l_deferredQueueSto, Q_DIM(l_deferredQueueSto), "deferred", &me->super);
    /* initialize extended state variables */
    me->config = MMI_CONFIG_OK;
    me->ds485Stopped = FALSE;
    me->datamodelStopped = FALSE;
    me->terminate = FALSE;
}

uint8_t MMI_GetDeferredQueueMargin(void)
{
    return QEQueue_getNFree(&l_mmi.deferredQueue);
}

/*..........................................................................*/
static QState MMI_initial(MMI *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_mmi);

    QS_FUN_DICTIONARY(&MMI_top);

    QS_FUN_DICTIONARY(&MMI_active);

    QS_FUN_DICTIONARY(&MMI_shutdown);
    QS_FUN_DICTIONARY(&MMI_invalid_configuration);
    QS_FUN_DICTIONARY(&MMI_ready);

    return Q_TRAN(&MMI_startup);
}

/*..........................................................................*/
static QState MMI_top(MMI *me, QEvent const *e)
{
    switch (e->sig) {
    case MMI_SHUTDOWN_SIG: {
        return Q_TRAN(&MMI_shutdown);
    }
    case DATAMODEL_STOPPED_SIG: {
        me->datamodelStopped = TRUE;
        if (me->terminate) {
            if (me->ds485Stopped) {
                MMI_Shutdown();
            }
        }
        return Q_HANDLED();
    }
    case DS485_STOPPED_SIG: {
        me->ds485Stopped = TRUE;
        if (me->terminate) {
            if (me->datamodelStopped) {
                MMI_Shutdown();
            }
        }
        return Q_HANDLED();
    }
    case SYSTEM_TERM_SIG: {
        me->terminate = TRUE;
        return Q_TRAN(&MMI_shutdown);
    }
    case CONFIG_INIT_SIG: {
        return Q_HANDLED();
    }
    case MMI_REQ_STATE_SIG: {
        MMIResStateEvt * resEvt = Q_NEW(MMIResStateEvt, MMI_RES_STATE_SIG);
        resEvt->hardware = MMI_HARDWARE_OK;
        resEvt->temperature = MMI_TEMPERATURE_OK;
        resEvt->config = me->config;
        resEvt->ds485 = MMI_DS485_READY;
        resEvt->line = MMI_LINE_IDLE;
        QActive_postFIFO(((MMIReqStateEvt const *)e)->requester, (QEvent *)resEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState MMI_active(MMI *me, QEvent const *e)
{
    switch (e->sig) {
    case MMI_CONFIG_STATE_SIG: {
        me->config = ((MMIConfigStateEvt const *)e)->config;
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&MMI_top);
}

/*..........................................................................*/
static QState MMI_startup(MMI *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* recall queued event */
        QActive_recall((QActive *)me, &me->deferredQueue);
        /* subscribe events */
        QActive_subscribe((QActive *)me, SYSTEM_TERM_SIG);
        QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
        return Q_HANDLED();
    }
    case MMI_CONFIG_STATE_SIG: {
        me->config = ((MMIConfigStateEvt * const)e)->config;
        if (me->config == MMI_CONFIG_OK) {
            return Q_TRAN(&MMI_ready);
        }
        /* configuration is invalid */
        return Q_TRAN(&MMI_invalid_configuration);
    }
    }
    return Q_SUPER(&MMI_top);
}

/*..........................................................................*/
static QState MMI_shutdown(MMI *me, QEvent const *e)
{
    return Q_SUPER(&MMI_top);
}

/*..........................................................................*/
static QState MMI_invalid_configuration(MMI *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* recall queued event */
        QActive_recall((QActive *)me, &me->deferredQueue);
        return Q_HANDLED();
    }
    case MMI_CONFIG_STATE_SIG: {
        me->config = ((MMIConfigStateEvt const *)e)->config;
        if (me->config == MMI_CONFIG_OK) {
            return Q_TRAN(MMI_DetermineState(me));
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&MMI_active);
}

/*..........................................................................*/
static QState MMI_ready(MMI *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* recall queued event */
        QActive_recall((QActive *)me, &me->deferredQueue);
        return Q_HANDLED();
    }
    case MMI_CONFIG_STATE_SIG: {
        me->config = ((MMIConfigStateEvt const *)e)->config;
        if (me->config != MMI_CONFIG_OK) {
            return Q_TRAN(&MMI_invalid_configuration);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&MMI_active);
}

/*..........................................................................*/
static QStateHandler MMI_DetermineState(MMI *me)
{
    /* priority 3 */
    if (me->config != MMI_CONFIG_OK) {
        return (QStateHandler)(&MMI_invalid_configuration);
    }
    /* priority 5 */
    return(QStateHandler)(&MMI_ready);
}

/*..........................................................................*/
static void MMI_Shutdown(void)
{
    BSP_SystemReset();
}

/*..........................................................................*/
static void MMI_SendChangeEvent(MMI *me)
{
    DsmApi_SendStatusEvent(MMI_HARDWARE_OK,
                           MMI_TEMPERATURE_OK,
                           me->config,
                           MMI_DS485_READY,
                           MMI_LINE_IDLE);
}
