/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of libdSvDC.

    libproputil is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    libproputil is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "msgutil.h"
#include "proputils.h"
#include "util.h"

int proputil_set_property_add(Vdcapi__VdsmRequestSetProperty *property,
                                 Vdcapi__PropertyElement *element)
{
    if (!property) {
        vdsm_report(LOG_ERR, "invalid property handle\n");
        return -1;
    }

    if (!property->n_properties) {
        property->properties = malloc (sizeof(Vdcapi__PropertyElement *));
        if (!property->properties) {
            vdsm_report(LOG_ERR, "could not allocate memory for property\n");
            return -1;
        }
    } else {
        Vdcapi__PropertyElement *properties;
        properties = realloc(property->properties,
                             sizeof(Vdcapi__PropertyElement *) * property->n_properties+1);
        if (!properties) {
            vdsm_report(LOG_ERR, "could not allocate memory for additional "
                            "properties\n");
            return -1;
        }
        property->properties = &properties;
    }

    property->properties[property->n_properties] = element;
    property->n_properties += 1;

    return 0;
}

static int proputil_property_prepare_element(
                                    const char *key,
                                    Vdcapi__PropertyElement **element)
{
    *element = NULL;

    Vdcapi__PropertyElement *el = malloc(sizeof(Vdcapi__PropertyElement));
    if (!el) {
        vdsm_report(LOG_ERR, "could not allocate property element\n");
        return -1;
    }
    vdcapi__property_element__init(el);

    el->name = strdup(key);
    if (!el->name) {
        vdsm_report(LOG_ERR, "could not allocate memory for property name\n");
        free(el);
        return -1;
    }

    Vdcapi__PropertyValue *val = malloc(sizeof(Vdcapi__PropertyValue));
    if (!val) {
        vdsm_report(LOG_ERR, "could not allocate memory for property value\n");
        free(el->name);
        free(el);
        return -1;
    }
    vdcapi__property_value__init(val);

    el->value = val;

    *element = el;

    return 0;
}

void proputil_set_property_free(Vdcapi__VdsmRequestSetProperty *property)
{
    if (!property) {
        return;
    }

    if (property->dsuid) {
        free(property->dsuid);
    }

    size_t i;
    for (i = 0; i < property->n_properties; ++i) {
        proputil_property_recursive_free(property->properties[i]);
    }

    if(property->n_properties) {
        free (property->properties);
    }

    free(property);
}

int proputil_set_property_add_uint(Vdcapi__VdsmRequestSetProperty *property,
                               const char *key,
                               uint64_t value)
{
    int ret;
    Vdcapi__PropertyElement *element = NULL;

    ret = proputil_property_prepare_element(key, &element);
    if (ret != 0) {
        return ret;
    }

    element->value->v_uint64 = value;
    element->value->has_v_uint64 = 1;

    ret = proputil_set_property_add(property, element);
    if (ret != 0) {
        free(element->value);
        free(element->name);
        free(element);
        return ret;
    }

    return 0;
}

int proputil_set_property_add_bool(Vdcapi__VdsmRequestSetProperty *property,
                               const char *key,
                               bool value)
{
    int ret;
    Vdcapi__PropertyElement *element = NULL;

    ret = proputil_property_prepare_element(key, &element);
    if (ret != 0) {
        return ret;
    }

    element->value->v_bool = value;
    element->value->has_v_bool = 1;

    ret = proputil_set_property_add(property, element);
    if (ret != 0) {
        free(element->value);
        free(element->name);
        free(element);
        return ret;
    }

    return 0;
}

int proputil_set_property_add_string(Vdcapi__VdsmRequestSetProperty *property,
                                 const char *key,
                                 const char *value,
                                 Vdcapi__PropertyElement **propElem)
{
    int ret;
    Vdcapi__PropertyElement *element = NULL;

    *propElem = NULL;

    if (value == NULL) {
        vdsm_report(LOG_ERR, "invalid string value\n");
        return -1;
    }

    ret = proputil_property_prepare_element(key, &element);
    if (ret != 0) {
        return ret;
    }

    element->value->v_string = strdup(value);
    if (!element->value->v_string) {
        vdsm_report(LOG_ERR, "could not allocate memory for string value\n");
        free(element->value);
        free(element->name);
        free(element);
        return -1;
    }

    ret = proputil_set_property_add(property, element);
    if (ret != 0) {
        free(element->value->v_string);
        free(element->value);
        free(element->name);
        free(element);
        return ret;
    }

    *propElem = element;
    return 0;
}

//**----------------------------------------------------------------------------
int proputil_get_property_new(const char* dsuid,
                              Vdcapi__VdsmRequestGetProperty **property)
{
    *property = NULL;

    if (!dsuid) {
        return -1;
    }

    Vdcapi__VdsmRequestGetProperty *p = malloc(sizeof(Vdcapi__VdsmRequestGetProperty));
    if (!p) {
        vdsm_report(LOG_ERR, "could not allocate new property\n");
        return -1;
    }

    vdcapi__vdsm__request_get_property__init(p);

    p->dsuid = strdup(dsuid);
    if (!p->dsuid) {
        vdsm_report(LOG_ERR, "could not allocate memory for dsuid name\n");
        free(p);
        return -1;
    }

    *property = p;
    return 0;
}

void proputil_get_property_free(Vdcapi__VdsmRequestGetProperty *property)
{
    if (property->dsuid) {
        free(property->dsuid);
    }

    size_t i;
    for (i = 0; i < property->n_query; ++i) {
        proputil_property_recursive_free(property->query[i]);
    }

    if (property->n_query > 0) {
        free(property->query);
    }

    free(property);
}

int proputil_get_property_add(Vdcapi__VdsmRequestGetProperty *property,
                              Vdcapi__PropertyElement *element)
{
    if (!property) {
        vdsm_report(LOG_ERR, "invalid property handle\n");
        return -1;
    }

    if (!property->n_query) {
        property->query = malloc (sizeof(Vdcapi__PropertyElement *));
        if (!property->query) {
            vdsm_report(LOG_ERR, "could not allocate memory for property\n");
            return -1;
        }
    } else {
        Vdcapi__PropertyElement *properties;
        properties = realloc(property->query,
                             sizeof(Vdcapi__PropertyElement *) * property->n_query+1);
        if (!properties) {
            vdsm_report(LOG_ERR, "could not allocate memory for additional "
                            "properties\n");
            return -1;
        }
        property->query = &properties;
    }

    property->query[property->n_query] = element;
    property->n_query += 1;

    return 0;
}


//**----------------------------------------------------------------------------
int proputil_set_property_new(const char *dsuid,
                              Vdcapi__VdsmRequestSetProperty **property)
{
    *property = NULL;

    if (!dsuid) {
        return -1;
    }

    Vdcapi__VdsmRequestSetProperty *p = malloc(sizeof(Vdcapi__VdsmRequestSetProperty));
    if (!p) {
        vdsm_report(LOG_ERR, "could not allocate new set property\n");
        return -1;
    }

    vdcapi__vdsm__request_set_property__init(p);

    p->dsuid = strdup(dsuid);
    if (!p->dsuid) {
        vdsm_report(LOG_ERR, "could not allocate memory for dsuid name\n");
        free(p);
        return -1;
    }

    *property = p;
    return 0;
}

//**----------------------------------------------------------------------------
int proputil_create_property_empty(Vdcapi__PropertyElement **element, const char* key)
{
    *element = NULL;

    Vdcapi__PropertyElement *el = malloc(sizeof(Vdcapi__PropertyElement));
    if (!el) {
        vdsm_report(LOG_ERR, "could not allocate property element\n");
        return -1;
    }
    vdcapi__property_element__init(el);

    el->name = strdup(key);
    if (!el->name) {
        vdsm_report(LOG_ERR, "could not allocate memory for property name\n");
        free(el);
        return -1;
    }

    *element = el;
    return 0;
}

//**----------------------------------------------------------------------------
int proputil_create_property_empty_index(Vdcapi__PropertyElement **element, int index)
{
    char buf[5];
    sprintf(buf, "%i", index);
    return proputil_create_property_empty(element, buf);
}

//**----------------------------------------------------------------------------
int proputil_create_property_int(Vdcapi__PropertyElement **element,
                                 const char *key,
                                 int64_t value)
{
    *element = NULL;
    Vdcapi__PropertyElement *el = NULL;
    if (proputil_property_prepare_element(key, &el) != 0) {
        return -1;
    }
    el->value->v_int64 = value;
    el->value->has_v_int64 = 1;

    *element = el;
    return 0;
}

int proputil_create_property_uint(Vdcapi__PropertyElement **element,
                                  const char *key,
                                  uint64_t value)
{
    *element = NULL;
    Vdcapi__PropertyElement *el = NULL;
    if (proputil_property_prepare_element(key, &el) != 0) {
        return -1;
    }
    el->value->v_uint64 = value;
    el->value->has_v_uint64 = 1;

    *element = el;
    return 0;
}

int proputil_create_property_bool(Vdcapi__PropertyElement **element,
                                  const char *key,
                                  bool value)
{
    *element = NULL;
    Vdcapi__PropertyElement *el = NULL;
    if (proputil_property_prepare_element(key, &el) != 0) {
        return -1;
    }
    el->value->v_bool = value;
    el->value->has_v_bool = 1;

    *element = el;
    return 0;
}

int proputil_create_property_double(Vdcapi__PropertyElement **element,
                                    const char *key,
                                    double value)
{
    *element = NULL;
    Vdcapi__PropertyElement *el = NULL;
    if (proputil_property_prepare_element(key, &el) != 0) {
        return -1;
    }
    el->value->v_double = value;
    el->value->has_v_double = 1;

    *element = el;
    return 0;
}

int proputil_create_property_string(Vdcapi__PropertyElement **element,
                                    const char *key,
                                    const char *value)
{
    *element = NULL;
    Vdcapi__PropertyElement *el = NULL;
    if (proputil_property_prepare_element(key, &el) != 0) {
        return -1;
    }
    el->value->v_string = strdup(value);
    if (!el->value->v_string) {
         vdsm_report(LOG_ERR, "could not allocate memory for property string\n");
        free(el->value);
        free(el);
        return -1;
    }
    el->value->has_v_bytes = 1;

    *element = el;
    return 0;
}


//**----------------------------------------------------------------------------
int proputil_get_property_find_element_by_name(Vdcapi__VdcResponseGetProperty *reqGetProperty,
                                                char *name,
                                                Vdcapi__PropertyElement **out)
{
    if (!reqGetProperty || !name || !out)  return FALSE;

    size_t i;
    for (i = 0; i < reqGetProperty->n_properties; ++i) {

        if ((reqGetProperty->properties[i]) &&
            (reqGetProperty->properties[i]->name) &&
            (strcmp(reqGetProperty->properties[i]->name, name) == 0)) {
            *out = reqGetProperty->properties[i];
            return 0;
        }
    }
    return -1;
}

//**----------------------------------------------------------------------------
int proputil_send_push_property_find_element_by_name(Vdcapi__VdcSendPushNotification * pushProperty,
                                                     char *name,
                                                     Vdcapi__PropertyElement **out)
{
    if (!pushProperty || !name || !out)  return FALSE;

    size_t i;
    for (i = 0; i < pushProperty->n_changedproperties; ++i) {

        if ((pushProperty->changedproperties[i]) &&
            (pushProperty->changedproperties[i]->name) &&
            (strcmp(pushProperty->changedproperties[i]->name, name) == 0)) {
            *out = pushProperty->changedproperties[i];
            return 0;
        }
    }
    return -1;
}

int proputil_property_array_find_element_by_name(Vdcapi__PropertyElement **properties,
                                         int numOfElements,
                                         char *name,
                                         Vdcapi__PropertyElement **out)
{
    int i;
    for (i = 0; i < numOfElements; ++i) {
        if (strcmp(properties[i]->name, name) == 0) {
            *out = properties[i];
            return 0;
        }
    }
    return -1;
}

int proputil_find_property_element_by_name(Vdcapi__PropertyElement *property,
                                   char *name,
                                   Vdcapi__PropertyElement **out)
{
    if (!property || !name || !out)  return FALSE;
    size_t i;
    for (i = 0; i < property->n_elements; ++i) {

        *out = property->elements[i]; // temp
        if ((property->elements[i]) &&
            (property->elements[i]->name) &&
            strcmp(property->elements[i]->name, name) == 0) {
            *out = property->elements[i];
            return 0;
        }
    }
    return -1;
}

int proputil_find_property_element_by_index(Vdcapi__PropertyElement *property,
                                          int index,
                                          Vdcapi__PropertyElement **out)
{
    char buf[5];
    memset(buf, 0, 5);

    sprintf(buf, "%i", index);
    return proputil_find_property_element_by_name(property,buf, out);
}


int proputil_find_property_element_by_name_index(Vdcapi__PropertyElement *property,
                                          char *name,
                                          int index,
                                          Vdcapi__PropertyElement **out)
{
    if ((!property) || (!name) ||(!out)) return -1;

    Vdcapi__PropertyElement *elemTemp = NULL;
    Vdcapi__PropertyElement *elemOut = NULL;

    if (proputil_find_property_element_by_name(property, name, &elemTemp) != 0) {
        return -1;
    }

    if (proputil_find_property_element_by_index(elemTemp, index, &elemOut) != 0) {
        return -1;
    }

    *out = elemOut;
    return 0;
}



int proputil_add_property_element(Vdcapi__PropertyElement *pParent,
                                  Vdcapi__PropertyElement *pElement)
{
    if ((!pParent) || !(pElement)) {
        return -1;
    }

    if (!pParent->n_elements) {
        pParent->elements = malloc(sizeof(Vdcapi__PropertyElement*));
        if (!pParent->elements) {
            return -1;
        }
    } else {
        Vdcapi__PropertyElement** temp = pParent->elements;
        int nr_of_elems = pParent->n_elements+1;
        if(!(temp = realloc(temp,
                            (nr_of_elems) * sizeof(Vdcapi__PropertyElement*)))) {
            return -1;
        }
        (pParent->elements) = temp;
    }

    pParent->elements[pParent->n_elements] = pElement;
    pParent->n_elements+=1;
    return 0;
}

void proputil_property_recursive_free(Vdcapi__PropertyElement *pElement)
{
    if (!pElement) {
        return;
    }

    if(pElement->n_elements > 0) {
        size_t index;
        for (index = 0; index < pElement->n_elements; ++index) {
            proputil_property_recursive_free(pElement->elements[index]);
        }
        free (pElement->elements);
        pElement->n_elements = 0;
    }
    if (pElement->value) {
        free (pElement->value->v_string);
    }
    free (pElement->value);
    free (pElement->name);
    free (pElement);
}

int get_property_element_name_bool(Vdcapi__PropertyElement *property,
                                   char *name,
                                   bool *out)
{
    if ((!property) || (!name) || (!out)) return -1;
    Vdcapi__PropertyElement* temp;
    if (proputil_find_property_element_by_name(property, name, &temp) != 0) {
        return -1;
    }
    if (!temp->value) {
        return -1;
    }
    if (!msgutil_get_bool(temp->value, out)) {
        return -1;
    }
    return 0;
}

int get_property_element_name_int(Vdcapi__PropertyElement *property,
                                   char *name,
                                   int64_t *out)
{
    if ((!property) || (!name) || (!out)) return -1;
    Vdcapi__PropertyElement* temp;
    if (proputil_find_property_element_by_name(property, name, &temp) != 0) {
        return -1;
    }
    if (!temp->value) {
        return -1;
    }
    if (!msgutil_get_int(temp->value, out)) {
        return -1;
    }
    return 0;
}

int get_property_element_name_uint(Vdcapi__PropertyElement *property,
                                   char *name,
                                   uint64_t *out)
{
    if ((!property) || (!name) || (!out)) return -1;
    Vdcapi__PropertyElement* temp;
    if (proputil_find_property_element_by_name(property, name, &temp) != 0) {
        return -1;
    }
    if (!temp->value) {
        return -1;
    }
    if (!msgutil_get_uint(temp->value, out)) {
        return -1;
    }
    return 0;
}

int get_property_element_name_double(Vdcapi__PropertyElement *property,
                                   char *name,
                                   double *out)
{
    if ((!property) || (!name) || (!out)) return -1;
    Vdcapi__PropertyElement* temp;
    if (proputil_find_property_element_by_name(property, name, &temp) != 0) {
        return -1;
    }
    if (!temp->value) {
        return -1;
    }
    if (!msgutil_get_double(temp->value, out)) {
        return -1;
    }
    return 0;
}

int get_property_element_name_string(Vdcapi__PropertyElement *property,
                                   char *name,
                                   const char **out)
{
    if ((!property) || (!name) || (!out)) return -1;
    Vdcapi__PropertyElement* temp;
    if (proputil_find_property_element_by_name(property, name, &temp) != 0) {
        return -1;
    }
    if (!temp->value) {
        return -1;
    }
    if (!msgutil_get_string(temp->value, out)) {
        return -1;
    }
    return 0;
}



